{{table['script'] | safe}}
<div class="abc" style="float: left; margin-left: 15%; margin-right:15%; width: 70%" >
    <div class="box">
        <h2>{{table['title']}}</h2>
        <p>{{table['help'] | safe}}</p>
        <div style=" margin-left: 3%; margin-right:3%;">
            {{table['df'] | safe}}
        </div>
    </div>
</div>