{% include "html/utils/start.html"%}

    {% if demultiplexing_summary is defined %}
    {% set step_summary = demultiplexing_summary %}
    {% include "html/utils/step.html"%}
    {% endif %}

    {% if mapping_summary is defined %}
    {% set step_summary = mapping_summary %}
    {% include "html/utils/step.html" %}
    {% endif %}

    {% if count_pathseq_summary is defined %}
    {% set step_summary = count_pathseq_summary %}
    {% include "html/utils/step.html" %}
    {% endif %}

    {% if analysis_pathseq_summary is defined %}
    {% set step_summary = analysis_pathseq_summary %}
    {% include "html/pathseq/analysis_pathseq.html"%}
    {% endif %}
        
{% include "html/utils/end.html"%}